#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
/***************************************************************************
 GenerateDHM
                                 A QGIS plugin
 From a DEM and a DSM generate a DHM by substracting the DSM to the DEM.
 Resolution of either the DEM or the DSM can be chosen.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2024-01-05
        copyright            : (C) 2022 by Gabriel Marques / Dynafor
        email                : gabriel.marques@toulouse-inp.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = "Gabriel Marques / Dynafor"
__date__ = "2024-01-05"
__copyright__ = "(C) 2024 by Gabriel Marques / Dynafor"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (QgsProcessing,
                       QgsProcessingAlgorithm,
                       QgsProcessingException,
                       QgsProcessingParameterMultipleLayers,
                       QgsProcessingParameterEnum,
                       QgsProcessingParameterBoolean,
                       QgsProcessingParameterNumber,
                       QgsProcessingParameterFolderDestination,
                       QgsRasterLayer,
                       QgsGeometry)

import uuid
import os
from typing import Union # until python 3.10 we need to use that instead of | 

from hedge_tools.utils.raster import utils as rutils
from hedge_tools.utils.raster import wrapper as wr
from hedge_tools.utils import utils as gutils
from qgis.PyQt.QtGui import QIcon
from hedge_tools import resources # Only need in hedge_tools.py normaly but just to keep track of import 

class GenerateDHMProcessingAlgorithm(QgsProcessingAlgorithm):
    """
    From a DEM and a DSM generate a DHM by substracting the DSM to the DEM.
    Resolution of either the DEM or the DSM can be chosen.

    Parameters
    ---

    Return
    ---
    
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    DSM = "DSM"
    DEM = "DEM"
    RES = "RES"
    RESAMPLING = "RESAMPLING"
    MASK_DHM = "MASK_DHM"
    THRESHOLD_MASK = "THRESHOLD_MASK"
    KEEP_INTERM = "KEEP_INTERM"
    OVERWRITE = "OVERWRITE"
    OUTPUT_DIR = "OUTPUT_DIR"

    def initAlgorithm(self, config=None):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.DSM,
                self.tr("DSM"),
                QgsProcessing.TypeRaster
            )
        )

        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.DEM,
                self.tr("DEM"),
                QgsProcessing.TypeRaster
            )
        )
            
        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.RES,
                description=self.tr("Resolution source for the DHM"),
                options=["DSM","DEM"],
                optional=False,
                allowMultiple=False
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.RESAMPLING,
                description=self.tr("Resampling method"),
                options=["Nearest neighbour ", "Bilinear", 
                         "Cubic", "Cubic spline ", 
                         "Lanczos windowed sinc", "Average", "Mode", 
                         "Maximum", "Minimum", "Median",
                         "First quartile", "Third quartile"],
                optional=False,
                allowMultiple=False,
                defaultValue=5
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.MASK_DHM,
                self.tr("Compute DHM masks (required for tree cover tool)"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.THRESHOLD_MASK,
                description=self.tr("Threshold height value (DHM masks)"),
                type=QgsProcessingParameterNumber.Double,
                defaultValue=2.5,
                optional=True,
                minValue=-999,
                maxValue=1000
            )
        )
        
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.KEEP_INTERM,
                self.tr("Keep intermediate files (will create sub directories within input directories)")
            )
        )
        
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.OVERWRITE,
                self.tr("Overwrite"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterFolderDestination(
                self.OUTPUT_DIR,
                self.tr("Output DHM directory")
            )
        )

    def __init__(self):
        """
        """
        super().__init__()

        self.just_created = []

    def processAlgorithm(self, parameters, context, feedback):
        """
        Here is where the processing itself takes place.
        """
        dsm = self.parameterAsLayerList(parameters, self.DSM, context)
        dem =  self.parameterAsLayerList(parameters, self.DEM, context)
        dem_res = self.parameterAsEnum(parameters, self.RES, context) # Used as a bool
        resampling = self.parameterAsEnum(parameters, self.RESAMPLING, context)
        compute_mask = self.parameterAsBoolean(parameters, self.MASK_DHM, context)
        threshold = self.parameterAsDouble(parameters, self.THRESHOLD_MASK, context)
        keep_interm_files = self.parameterAsBoolean(parameters, self.KEEP_INTERM, context)
        overwrite=  self.parameterAsBoolean(parameters, self.OVERWRITE, context)
        out_dir = self.parameterAsFileOutput(parameters, self.OUTPUT_DIR, context)

        assign_proj = False # [is it necessary to allow this option ?]

        feedback.pushInfo("Handling rasters resolution and relative size.")
        # Check for cancellation
        if feedback.isCanceled():
            return {}
        
        # Fetch resolution for DHM
        resolution = self.get_resolution_value(dem_res, dem[0], dsm[0])
        # Fetch max resolution between source to ensure common_extent is bigger than max pixel size
        # Assuming square pixels
        max_res = max(dem[0].rasterUnitsPerPixelX(), dsm[0].rasterUnitsPerPixelX())
        # Fetch biggest tile to iterate on and fetch intersecting smaller tiles
        biggest = self.fetch_biggest_tiles(dsm, dem)

        # Check for cancellation
        if feedback.isCanceled():
            return {}
        
        # Define which of the DEM or DSM have the bigger tile 
        # and handle changes in workspace and formula accordingly
        rasters, tiles, formula, workspace_raster, workspace_tile = \
            self.handle_iteration(dsm, dem, biggest, keep_interm_files, dem_res)
        if compute_mask:
            gutils.create_folders(out_dir, ["dhm_masks"])

        # Step for progress bar
        # Count is an estimation because there 
        # will be tiles that are not intersecting
        count = len(tiles) 
        current = 0

        feedback.pushInfo("Creating DHM.")
        for raster in rasters:
            part = 1 # To name raster parts when clip by tile
            if isinstance(raster, str):
                raster = rutils.load_raster(raster)
            # Fetch rasters intersecting biggest raster
            intersect = rutils.fetch_intersecting_tiles(raster, tiles)
            
            if feedback.isCanceled():
                return {}
            
            for tile in intersect:
                # Assign projection [is it necessary to allow this option ?]
                if assign_proj:
                    wr.assign_projection(raster)

                # Get common extent for clipping raster and align pixels
                common_ext = rutils.get_common_extent(raster, tile)
                bb = common_ext.boundingBox()
                # Ignore tiles where common_ext < max pixel resolution
                if (bb.xMaximum() - bb.xMinimum()) >= max_res \
                    and (bb.yMaximum() - bb.yMinimum()) >= max_res:

                    # Clip and align raster and tile
                    warp, clip = \
                        self.dhm_preprocess(raster, tile, resolution, 
                                            common_ext, resampling, biggest, dem_res, 
                                            keep_interm_files, overwrite,
                                            workspace_raster, workspace_tile, 
                                            part)  

                    if feedback.isCanceled():
                        return {}
                
                    tile_name = tile.name()
                    # Compute dhm and masks if wanted
                    self.generate_dhm(warp, clip, formula, dem_res, 
                                      compute_mask, threshold, overwrite,
                                      out_dir, tile_name, part)
                    current += 1
                    part += 1

                # Set progress
                feedback.setProgress(int((current/count) * 100))
                # Check for cancellation
                if feedback.isCanceled():
                    return {}

        return {self.OUTPUT_DIR: out_dir}

    def handle_iteration(self, dsm:list[QgsRasterLayer], dem:list[QgsRasterLayer], 
                         biggest:bool, keep_interm_files:bool, dem_res:bool
                         )->tuple[list[QgsRasterLayer], list[QgsRasterLayer],
                                  str, str, str]:
        """
        Define which of the DEM or DSM as the biggest tiles for future iteration.
        Define the DHM generation formula aswell depending of which one as the bigger tile.
        If intermediary files are kept, create subfolders within DEM and DSM directory.

        Parameters
        ---
        dsm:list[QgsRasterLayer]
            List of dsm raster
        dem:list[QgsRasterLayer]
            List of dem raster
        biggest:bool
            If True a DSM tile is bigger than a DEM tile
        keep_interm_files:bool
            If True subfolders will be created within parent directory of DEM and DSM
        dem_res:bool
            If True the DEM resolution will be kept, which modify treatment order.
            This change is echoed on subfolders creation

        Return
        ---
        rasters: list[QgsRasterLayer]
            List of the bigger tiles, either DSM or DEM
        tiles: list[QgsRasterLayer]
            List of the smaller tiles, either DSM or DEM
        formula: str
            Formula for DHM generation. Dependant of which one as the bigger tile
        workspace_raster: str
            Directory of the bigger tiles
        workspace_tile: str
            Directory of the smaller tiles
        """
        if biggest:
            rasters = dsm
            tiles = dem
            formula = "B-A"       
            if keep_interm_files and dem_res:
                workspace_raster = self.create_interm_folders(dsm[0], ["clipped"])
                workspace_tile = self.create_interm_folders(dem[0], ["aligned"])
            if keep_interm_files and dem_res == False:
                workspace_raster = self.create_interm_folders(dsm[0], ["aligned"])
                workspace_tile = self.create_interm_folders(dem[0], ["clipped"])
        else:
            rasters = dem
            tiles = dsm
            formula = "A-B"
            if keep_interm_files and dem_res:
                workspace_tile = self.create_interm_folders(dsm[0], ["aligned"])
                workspace_raster = self.create_interm_folders(dem[0], ["clipped"])
            if keep_interm_files and dem_res == False:
                workspace_tile = self.create_interm_folders(dsm[0], ["clipped"])
                workspace_raster = self.create_interm_folders(dem[0], ["aligned"])
        
        if keep_interm_files is False:
            workspace_raster = None
            workspace_tile = None
            
        return rasters, tiles, formula, workspace_raster, workspace_tile
    
    def create_interm_folders(self, file:Union[QgsRasterLayer, str], folders:list[str])->str:
        """
        Create folders in a parent file directory.

        Parameters
        ---
        input_file: QgsRasterLayer | str
            Path of a file to fetch the parent directory from.
        folders: list[str]
            Name of the folders that'll be created in the parent directory
        
        Return
        ---
        workspace: str
            Return the path of the parent directory for convenience
        """
        if isinstance(file, QgsRasterLayer):
            path = file.dataProvider().dataSourceUri()
        else:
            path = file

        workspace, _ = os.path.split(path)
        gutils.create_folders(workspace, folders)

        return workspace

    def get_resolution_value(self, dem_res:bool, 
                             dem:Union[QgsRasterLayer, str], 
                             dsm:Union[QgsRasterLayer, str])->float:
        """
        From the value of a boolean parameter determine 
        if the dem resolution or the dsm resolution should be taken
        
        Parameters
        ---
        dem_res: boolean
            If True DEM resolution will be used
            Else DSM resolution will be used
        dem: QgsRasterLayer | str
        dsm: QgsRasterLayer | str
        
        Return
        ---
        resolution:float
        """
        if isinstance(dem, str):
            dem = rutils.load_raster(dem)
        if isinstance(dsm, str):
            dsm = rutils.load_raster(dsm)
        
        # Assuming squared pixels
        # Round because QGIS will sometimes create non null decimal 
        # at 17th or 18th place which result in missalignement
        if dem_res:
            resolution = round(dem.rasterUnitsPerPixelX(), 4)
            resolution_2 = round(dsm.rasterUnitsPerPixelX(), 4)
        else:
            resolution = round(dsm.rasterUnitsPerPixelX(), 4)
            resolution_2 = round(dem.rasterUnitsPerPixelX(), 4)

        return resolution
        
    def fetch_biggest_tiles(self, 
                            rasters_1:list[Union[QgsRasterLayer, str]], 
                            rasters_2:list[Union[QgsRasterLayer, str]])->bool:
        """
        Compare area of two raster and return a boolean.
        
        Parameters:
        ---
        rasters_1: list[QgsRasterLayer | Str]
        rasters_2: list[QgsRasterLayer | Str]
        
        Return
        ---
        biggest:bool
            True if raster_1 > raster_2 else False
        """ 
        biggest = False
        
        max_area_1 = 0
        for raster_1 in rasters_1:
            if isinstance(raster_1, str):
                raster_1 = rutils.load_raster(raster_1)
            if raster_1.extent().area() > max_area_1:
                max_area_1 = raster_1.extent().area() 
                    
        max_area_2 = 0
        for raster_2 in rasters_2:
            if isinstance(raster_2, str):
                raster_2 = rutils.load_raster(raster_2)
            if raster_2.extent().area() > max_area_2:
                max_area_2 = raster_2.extent().area() 
                    
        
        if max_area_1 > max_area_2:
            biggest = True
        
        return biggest
    
    def check_files_exists(self, files:list[str], overwrite:bool)->None:
        """
        From a list of path, check if a files already exists.
        If owerwrite is not toggle, algorithm will end there 
        and a feedback will be pushed to the user.

        Parameters
        ---
        files: list[str]
            Path of files to check
        overwrite: bool
            Overwrite parameter of the algorithm
        Return
        ---
        None
        """
        for file in files:
            exists = gutils.is_file_exists(file)
            if exists and overwrite is False:
                _, tail = os.path.split(file)
                raise QgsProcessingException(
                    f"{tail} already exists. Please delete files/folders \
                    or toggle overwrite parameter")
            if exists and overwrite and gutils.is_file_exists(f"{file}.aux.xml"):
                os.remove(f"{file}.aux.xml")
    
    def dhm_preprocess(self, raster:QgsRasterLayer, tile:QgsRasterLayer, 
                       resolution:float, common_ext:QgsGeometry, resampling:int,
                       biggest:bool, dem_res:bool, keep_interm_files:bool, 
                       overwrite:bool, workspace_raster:str, workspace_tile:str, 
                       count:int)->tuple[str, str]:
        """
        Align and clip both DEM and DSM accordingly with DHM wanted resolution.
        If intermediary files should be kept 
        it'll write the various outputs in DEM or DSM directory.

        Some case inside the method seems redo

        Parameters
        ---
        raster: QgsRasterLayer
            Either DEM or DSM tile depending of which as the bigger ones.
        tile: QgsRasterLayer 
            Either DEM or DSM tile depending of which as the smaller ones.
        resolution: float
            Resampling resolution for DEM or DSM raster.
        common_ext: QgsGeometry
            Common extent between the DSM and DEM tiles.
        resampling: int
            Gdal warp algorithm to be used
        biggest: bool
            If True the biggest tile (raster) is the DSM.
        dem_res: bool
            If True DEM resolution will be applied to DSM.
            Else DSM resolution will be applied to DEM.
        keep_interm_files: bool
            If True clipped and warped rasters will be saved in DEM and DSM directory.
        overwrite: bool
            Will overwrite output if it exists
        workspace_raster: str
            Directory of either the DSM or DEM.
        workspace_tile: str
            Directory of either the DSM or DEM.
        count: int

        Return
        ---
        warp:str
            Warped DSM or DEM path.
        clip:str
            Clipped DSM or DEM path.
        """ 
        clip_uri = "TEMPORARY_OUTPUT"
        warp_uri = "TEMPORARY_OUTPUT"
        # Define if intermediary files should be kept and where to save them
        # Depending of resolution wanted by user
        if (biggest and dem_res) or (biggest == False and dem_res == False): 
            if keep_interm_files:
                # DHM = raster, NDVI = tile and DHM should be warp
                # DHM = tile, NDVI = raster and NDVI should be warp
                warp_uri = os.path.join(workspace_raster, "aligned", 
                                        f"{raster.name()}.tif")
                clip_uri = os.path.join(workspace_tile, "clipped", 
                                        f"{tile.name()}.tif")
                if clip_uri in self.just_created: # In case tile is between two rasters
                    clip_uri = os.path.join(workspace_tile, "clipped", 
                                            f"{tile.name()}_{count}.tif")
                if warp_uri in self.just_created:
                    warp_uri = os.path.join(workspace_raster, "aligned", 
                                            f"{raster.name()}_{count}.tif")
                self.check_files_exists([warp_uri, clip_uri], overwrite)

            # Align and clip raster between them    
            warp = wr.gdal_warp(raster, resolution, common_ext, resampling, warp_uri, overwrite)
            clip = wr.gdal_clip_by_extent(tile, common_ext, clip_uri)
        else:
            if keep_interm_files:
                # DHM = raster, NDVI = tile and NDVI should be warp
                # DHM = tile, NDVI = raster and DHM should be warp
                clip_uri = os.path.join(workspace_raster, "clipped", 
                                        f"{raster.name()}_{count}.tif")
                warp_uri = os.path.join(workspace_tile, "aligned", 
                                        f"{tile.name()}.tif")
                if clip_uri in self.just_created: # In case tile is between two rasters
                    clip_uri = os.path.join(workspace_raster, "clipped", 
                                            f"{raster.name()}_{count}.tif")
                if warp_uri in self.just_created:
                    warp_uri = os.path.join(workspace_tile, "aligned", 
                                            f"{tile.name()}_{count}.tif")
                self.check_files_exists([clip_uri, warp_uri], overwrite)

            # Align and clip raster between them
            warp = wr.gdal_warp(tile, resolution, common_ext, resampling, warp_uri, overwrite)
            clip = wr.gdal_clip_by_extent(raster, common_ext, clip_uri)

        return warp, clip
    
    def generate_dhm(self, warp:str, clip:str, formula:str, 
                     dem_res: bool, compute_mask:bool, threshold:float, 
                     overwrite:bool, out_dir:str, name:str, count:int)->None:
        """
        Generate DHM from a DEM and a DSM with the same extent and resolution.
        It is also possible to threshold the DHM 
        in this case a folder will be created in the DHM folder.

        Parameters
        ---
        warp:str
            Path of the warped raster. Either DEM or DSM.
        clip:str
            Path of the clipped raster. Either DEM or DSM.
        formula:str 
            Order of substraction depending of tile size between DEM and DSM.
        dem_res: bool
            If True DEM resolution will be applied to DSM.
            Else DSM resolution will be applied to DEM.
        compute_mask:bool
            If True will mask pixel value of the DHM below the threshold.
        threshold:float
            Value used to threshold the DHM.
        overwrite:bool
            Will overwrite output if it exists
        out_dir: str
            Directory where DHM will be saved.
            If the masks are created they'll be saved within a subdirectory.
        name: str
            Name of the input raster which will be used to name the created DHM
            -> DHM_name
        count: int
        
        Return
        ---
        None
        """
        dhm_uri = os.path.join(out_dir, f"DHM_fr_{name}.tif")
        if dhm_uri in self.just_created:
            dhm_uri = os.path.join(out_dir, f"DHM_fr_{name}_{count}.tif")
        self.check_files_exists([dhm_uri], overwrite)

        if dem_res:
            calc = wr.gdal_calc([(warp, 1), (clip, 1)], formula, 
                                output=dhm_uri, overwrite=overwrite)
        else:
            calc = wr.gdal_calc([(clip, 1), (warp, 1)], formula, 
                                output=dhm_uri, overwrite=overwrite)

        if compute_mask:
            mask_uri = os.path.join(out_dir, "dhm_masks", f"MASK_fr_{name}.tif")
            if mask_uri in self.just_created:
                mask_uri = os.path.join(out_dir, "dhm_masks", f"MASK_fr_{name}_{count}.tif")
            self.check_files_exists([mask_uri], overwrite)
            _ = wr.gdal_calc([(calc, 1)], f"where(A >= {threshold}, 1, 0)", 
                             0, output=mask_uri, overwrite=overwrite)
        
    def icon(self):
        """
        Should return a QIcon which is used for your provider inside
        the Processing toolbox.
        """
        return QIcon(":/plugins/hedge_tools/images/hedge_tools.png")

    def shortHelpString(self):
        """
        Returns a localised short helper string for the algorithm. This string
        should provide a basic description about what the algorithm does and 
        the parameters and outputs associated with it.
        """
        return self.tr("Generate a DHM from any intersecting DEM and DSM. \
                        Algorithm allow to select either the DEM or DSM resolution \
                        for the DHM. \n\
                        The user can also select the resampling method to use \
                        when adjusting the resolution. Some methods yield better results \
                        for downsampling while others are better for upsampling. \
                        Beware some methods (like nearest neighbor) might create space between tiles. \n\
                        The user have the option to threshold the generated DHM. \
                        This option is mandatory if the tree cover tools will be used. \n\
                        It is also possible to keep intermediary files \
                        such as clipped and warped DEM or DSM.")
    
    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "generatedhm"

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr("1 - Generate DHM")

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr("0 - Extraction [optional]")

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "extraction"
    
    def tr(self, string):
        """
        Returns a translatable string with the self.tr() function.
        """
        return QCoreApplication.translate("Processing", string)
    
    def checkParameterValues(self, parameters, context):
        
        if parameters[self.OUTPUT_DIR]  == "TEMPORARY_OUTPUT":
            return (False, "It is mandatory to provide a DHM output directory.")
        
        return (True, '')

    def createInstance(self):
        return GenerateDHMProcessingAlgorithm()
                    
