#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
/***************************************************************************
 GenerateDHM
                                 A QGIS plugin
 From a DEM and a DSM generate a DHM by substracting the DSM to the DEM.
 Resolution of either the DEM or the DSM can be chosen.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2024-01-05
        copyright            : (C) 2022 by Gabriel Marques / Dynafor
        email                : gabriel.marques@toulouse-inp.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = "Gabriel Marques / Dynafor"
__date__ = "2024-01-05"
__copyright__ = "(C) 2024 by Gabriel Marques / Dynafor"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (QgsProcessing,
                       QgsProcessingAlgorithm,
                       QgsProcessingException,
                       QgsProcessingParameterMultipleLayers,
                       QgsProcessingParameterEnum,
                       QgsProcessingParameterBoolean,
                       QgsProcessingParameterNumber,
                       QgsProcessingParameterExtent,
                       QgsProcessingParameterFolderDestination,
                       QgsFeedback)

import uuid
import os
from typing import Union # until python 3.10 we need to use that instead of | 

from hedge_tools.utils.raster import utils as rutils
from hedge_tools.utils.raster import wrapper as wr
from hedge_tools.utils import utils as gutils
from qgis.PyQt.QtGui import QIcon
from hedge_tools import resources # Only need in hedge_tools.py normaly but just to keep track of import 

class GenerateNDVIProcessingAlgorithm(QgsProcessingAlgorithm):
    """
    From a raster with infrared and red band generate a NDVI raster.
    En extent can be specified to cut the sources.

    Parameters
    ---

    Return
    ---
    
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    RASTER = "RASTER"
    EXTENT = "EXTENT"
    RED_BAND = "RED_BAND"
    IR_BAND = "IR_BAND"
    MASK_NDVI = "MASK_NDVI"
    THRESHOLD_MASK = "THRESHOLD_MASK"
    SIEVE = "SIEVE"
    THRESHOLD_SIEVE = "THRESHOLD_SIEVE"
    CONNECTEDNESS = "CONNECTEDNESS"
    OVERWRITE = "OVERWRITE"
    OUTPUT_DIR = "OUTPUT_DIR"

    def initAlgorithm(self, config=None):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.RASTER,
                self.tr("Input rasters"),
                QgsProcessing.TypeRaster
            )
        )

        self.addParameter(
            QgsProcessingParameterExtent(
                self.EXTENT,
                self.tr("Extent of the study area"),
                optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.RED_BAND,
                description=self.tr("Red band"),
                type=QgsProcessingParameterNumber.Integer,
                defaultValue=2,
                minValue=1,
                maxValue=100
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.IR_BAND,
                description=self.tr("Infrared band"),
                type=QgsProcessingParameterNumber.Integer,
                defaultValue=1,
                minValue=1,
                maxValue=100
            )
        )
            
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.MASK_NDVI,
                self.tr("Compute NDVI masks (required for tree cover tool)"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.THRESHOLD_MASK,
                description=self.tr("Threshold NDVI value (NDVI masks)"),
                type=QgsProcessingParameterNumber.Double,
                defaultValue=0.2,
                optional=True,
                minValue=-1,
                maxValue=1
            )
        )
        
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.SIEVE,
                self.tr("Compute sieved NDVI masks"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.THRESHOLD_SIEVE,
                description=self.tr("Threshold sieve value (pixel number)"),
                type=QgsProcessingParameterNumber.Integer,
                defaultValue=100,
                optional=True,
                minValue=0
            )
        )
        
        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.CONNECTEDNESS,
                description=self.tr("Connectedness parameter for sieved output"),
                options=["4","8"],
                defaultValue=1,
                optional=True,
                allowMultiple=False
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.OVERWRITE,
                self.tr("Overwrite"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterFolderDestination(
                self.OUTPUT_DIR,
                self.tr("Output NDVI directory")
            )
        )

    def processAlgorithm(self, parameters, context, feedback):
        """
        Here is where the processing itself takes place.
        """

        rasters = self.parameterAsLayerList(parameters, self.RASTER, context)
        extent = self.parameterAsExtent(parameters, self.EXTENT, context)
        red_band =  self.parameterAsInt(parameters, self.RED_BAND, context)
        ir_band =  self.parameterAsInt(parameters, self.IR_BAND, context)
        compute_mask = self.parameterAsBoolean(parameters, self.MASK_NDVI, context)
        threshold_mask = self.parameterAsDouble(parameters, self.THRESHOLD_MASK, context)
        compute_sieve = self.parameterAsBoolean(parameters, self.SIEVE, context)
        threshold_sieve = self.parameterAsDouble(parameters, self.THRESHOLD_SIEVE, context)
        connectedness =  self.parameterAsInt(parameters, self.CONNECTEDNESS, context)
        overwrite=  self.parameterAsBoolean(parameters, self.OVERWRITE, context)
        out_dir = self.parameterAsFileOutput(parameters, self.OUTPUT_DIR, context)

        feedback.pushInfo("Starting processing")
         # Set total for progress bar
        count = len(rasters)
        # Check for cancellation
        if feedback.isCanceled():
            return {}
        
        # Create subdir if necessary
        self.create_folders(out_dir, compute_mask, compute_sieve)
        # Cast as float 32 to avoid divide by 0
        formula = "(B.astype(float32) - A.astype(float32))/\
                   (B.astype(float32) + A.astype(float32))"

        for current, raster in enumerate(rasters):
            # Update progress bar here as a loop can take time
            feedback.setProgress(int((current/count) * 100))
            name = raster.name()
            if isinstance(raster, str):
                raster = rutils.load_raster(raster)
            if extent.area() != 0:
                # Might create missing pixel without warping to dhm first, let's see
                feedback.pushInfo(f"Clipping {name} by extent...")
                # Check for cancellation
                if feedback.isCanceled():
                    return {}
                common_ext = rutils.get_common_extent(raster, extent)
                raster = wr.gdal_clip_by_extent(raster, common_ext)
            # Compute NDVI
            feedback.pushInfo(f"Computing NDVI for {name}...")
            # Check for cancellation
            if feedback.isCanceled():
                return {}
            ndvi_uri = os.path.join(out_dir, f"NDVI_fr_{name}.tif")
            self.check_files_exists([ndvi_uri], overwrite)
            ndvi = wr.gdal_calc([(raster, red_band), (raster, ir_band)], 
                                 formula, output=ndvi_uri, overwrite=overwrite)
            # Compute mask if wanted
            if compute_mask:
                feedback.pushInfo(f"Computing NDVI mask for {name}...")
                # Check for cancellation
                if feedback.isCanceled():
                    return {}
                mask_uri = os.path.join(out_dir, "mask", f"mask_fr_{name}.tif") 
                self.check_files_exists([mask_uri], overwrite)
                mask = wr.gdal_calc([(ndvi, 1)], f"where(A >= {threshold_mask}, 1, 0)",
                                    0, output=mask_uri, overwrite=overwrite)
            # Sieve mask if wanted
            if compute_sieve:
                feedback.pushInfo(f"Sieving NDVI mask for {name}...")
                # Check for cancellation
                if feedback.isCanceled():
                    return {}
                sieve_uri = os.path.join(out_dir, "sieve", f"sieve_fr_{name}.tif") 
                self.check_files_exists([sieve_uri], overwrite)
                mask = wr.gdal_sieve(mask, threshold_sieve, connectedness, sieve_uri)

            # Set progress
            feedback.pushInfo("=================================")
            # Check for cancellation
            if feedback.isCanceled():
                return {}
            
        return {self.OUTPUT_DIR: out_dir}
            
    def create_folders(self, out_dir: str, mask:bool, sieve:bool)->None:
        """
        Create folders for mask and sieve ndvi in the output directory if necessary.

        Parameters
        ---
        out_dir : str
            Path of the output dirctory where the folders will be created.
        mask: bool
            Will create a mask directory if True.
        sieve: bool
            Will create a sieve directory if True.

        Return
        ---
        None
        """
        folders = []

        if mask:
            folders.append("mask")
        if sieve:
            folders.append("sieve")

        if bool(folders):
            gutils.create_folders(out_dir, folders)
    
    def check_files_exists(self, files:list[str], overwrite:bool)->None:
        """
        From a list of path, check if a files already exists.
        If owerwrite is not toggle, algorithm will end there 
        and a feedback will be pushed to the user.

        Parameters
        ---
        files: list[str]
            Path of files to check
        overwrite: bool
            Overwrite parameter of the algorithm
        Return
        ---
        None
        """
        for file in files:
            exists = gutils.is_file_exists(file)
            if exists and overwrite is False:
                _, tail = os.path.split(file)
                raise QgsProcessingException(
                    f"{tail} already exists. Please delete files/folders \
                    or toggle overwrite parameter")
            if exists and overwrite and gutils.is_file_exists(f"{file}.aux.xml"):
                os.remove(f"{file}.aux.xml")

    def icon(self):
        """
        Should return a QIcon which is used for your provider inside
        the Processing toolbox.
        """
        return QIcon(":/plugins/hedge_tools/images/hedge_tools.png")

    def shortHelpString(self):
        """
        Returns a localised short helper string for the algorithm. This string
        should provide a basic description about what the algorithm does and 
        the parameters and outputs associated with it..
        """
        return self.tr("Generate NDVI rasters from one or \
                       multiple rasters with a red band and an infrared band. \
                       The band number use gdal system, the first band is 1. \n\
                       Optionnaly user can specify an extent, in which case only \
                       the common extent between the rasters and the given extent \
                       will be used for the generation.\n\
                       User can threshold the NDVI to create a binary masks. \
                       This option is mandatory if the tree cover tools will be used.\n\
                       It is also possible and advised to use the sieve option to \
                       remove small objects (connected pixels) in the masks. \
                       The sieve value is in pixel number. \n\
                       The connectedness parameter specify if 8 neighbooring pixels \
                       (diagonals included) should be tested for connection or only 4.")
    
    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "generatendvi"

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr("1 - Generate NDVI")

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr("0 - Extraction [optional]")

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "extraction"
    
    def tr(self, string):
        """
        Returns a translatable string with the self.tr() function.
        """
        return QCoreApplication.translate("Processing", string)
    
    def checkParameterValues(self, parameters, context):
        
        if parameters[self.OUTPUT_DIR]  == "TEMPORARY_OUTPUT":
            return (False, "It is mandatory to provide a NDVI output directory.")
        
        if parameters[self.SIEVE] and not parameters[self.MASK_NDVI]:
            return (False, "NDVI masks must be toggle if you want to use the sieve option.")
        
        return (True, '')

    def createInstance(self):
        return GenerateNDVIProcessingAlgorithm()
                    
