#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
/***************************************************************************
 GenerateDHM
                                 A QGIS plugin
 From a DEM and a DSM generate a DHM by substracting the DSM to the DEM.
 Resolution of either the DEM or the DSM can be chosen.
 Generated by Plugin Builder: http://g-sherman.github.io/Qgis-Plugin-Builder/
                              -------------------
        begin                : 2024-01-05
        copyright            : (C) 2022 by Gabriel Marques / Dynafor
        email                : gabriel.marques@toulouse-inp.fr
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = "Gabriel Marques / Dynafor"
__date__ = "2024-01-05"
__copyright__ = "(C) 2024 by Gabriel Marques / Dynafor"

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = "$Format:%H$"

from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (QgsProcessing,
                       QgsProcessingAlgorithm,
                       QgsProcessingException,
                       QgsProcessingParameterMultipleLayers,
                       QgsProcessingParameterEnum,
                       QgsProcessingParameterBoolean,
                       QgsProcessingParameterNumber,
                       QgsProcessingParameterFolderDestination,
                       QgsRasterLayer,
                       QgsGeometry)

import uuid
import os
import glob

from typing import Union # until python 3.10 we need to use that instead of | 

from hedge_tools.utils.raster import utils as rutils
from hedge_tools.utils.raster import wrapper as wr
from hedge_tools.utils.vector import qgis_wrapper as qw
from hedge_tools.utils import utils as gutils
from qgis.PyQt.QtGui import QIcon
from hedge_tools import resources # Only need in hedge_tools.py normaly but just to keep track of import 

class GenerateTreeCoverProcessingAlgorithm(QgsProcessingAlgorithm):
    """
    Combine a thresholded DHM and a thresholded NDVI to create a tree cover raster.

    Parameters
    ---

    Return
    ---
    
    """

    # Constants used to refer to parameters and outputs. They will be
    # used when calling the algorithm from another algorithm, or when
    # calling from the QGIS console.

    DHM = "DHM"
    NDVI = "NDVI"
    RES = "RES"
    RESAMPLING = "RESAMPLING"
    SIEVE = "SIEVE"
    THRESHOLD_SIEVE = "THRESHOLD_SIEVE"
    CONNECTEDNESS = "CONNECTEDNESS"
    KEEP_INTERM = "KEEP_INTERM"
    POLYGONIZE = "POLYGONIZE"
    CLOSE = "CLOSE"
    OVERWRITE = "OVERWRITE"
    OUTPUT_DIR = "OUTPUT_DIR"

    def initAlgorithm(self, config=None):
        """
        Here we define the inputs and output of the algorithm, along
        with some other properties.
        """

        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.DHM,
                self.tr("DHM"),
                QgsProcessing.TypeRaster
            )
        )

        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.NDVI,
                self.tr("NDVI"),
                QgsProcessing.TypeRaster
            )
        )
            
        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.RES,
                description=self.tr("Resolution source for the tree cover"),
                options=["DHM","NDVI"],
                optional=False,
                allowMultiple=False
            )
        )

        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.RESAMPLING,
                description=self.tr("Resampling method"),
                options=["Nearest neighbour ", "Bilinear", 
                         "Cubic", "Cubic spline ", 
                         "Lanczos windowed sinc", "Average", "Mode", 
                         "Maximum", "Minimum", "Median",
                         "First quartile", "Third quartile"],
                optional=False,
                allowMultiple=False,
                defaultValue=5
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.SIEVE,
                self.tr("Sieve tree cover"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterNumber(
                name=self.THRESHOLD_SIEVE,
                description=self.tr("Threshold sieve value (pixel number)"),
                type=QgsProcessingParameterNumber.Integer,
                defaultValue=50,
                optional=True,
                minValue=0
            )
        )
        
        self.addParameter(
            QgsProcessingParameterEnum(
                name=self.CONNECTEDNESS,
                description=self.tr("Connectedness parameter for sieved output"),
                options=["4","8"],
                defaultValue=1,
                optional=True,
                allowMultiple=False
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.KEEP_INTERM,
                self.tr("Keep intermediate files (will create sub directories within input directories)")
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.POLYGONIZE,
                self.tr("Polygonize tree cover rasters as one vector layer"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.CLOSE,
                self.tr("Morphological closing of the vector layer"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.OVERWRITE,
                self.tr("Overwrite"),
                defaultValue=True
            )
        )

        self.addParameter(
            QgsProcessingParameterFolderDestination(
                self.OUTPUT_DIR,
                self.tr("Output tree cover directory")
            )
        )

    def __init__(self):
        """
        """
        super().__init__()

        self.just_created = []
        self.files = []

    def processAlgorithm(self, parameters, context, feedback):
        """
        Here is where the processing itself takes place.
        """
        dhm = self.parameterAsLayerList(parameters, self.DHM, context)
        ndvi =  self.parameterAsLayerList(parameters, self.NDVI, context)
        ndvi_res = self.parameterAsEnum(parameters, self.RES, context) # Used as a bool
        resampling = self.parameterAsEnum(parameters, self.RESAMPLING, context)
        compute_sieve = self.parameterAsBoolean(parameters, self.SIEVE, context)
        threshold_sieve = self.parameterAsDouble(parameters, self.THRESHOLD_SIEVE, context)
        connectedness =  self.parameterAsInt(parameters, self.CONNECTEDNESS, context)
        keep_interm_files = self.parameterAsBoolean(parameters, self.KEEP_INTERM, context)
        polygonize = self.parameterAsBoolean(parameters, self.POLYGONIZE, context)
        close = self.parameterAsBoolean(parameters, self.CLOSE, context)
        overwrite=  self.parameterAsBoolean(parameters, self.OVERWRITE, context)
        out_dir = self.parameterAsFileOutput(parameters, self.OUTPUT_DIR, context)

        # Multiplicator for progress bar
        multiplicator = 100

        assign_proj = False # [is it necessary to allow this option ?]

        feedback.pushInfo("Handling rasters resolution and relative size.")
        # Check for cancellation
        if feedback.isCanceled():
            return {}
        
        # Fetch resolution for DHM
        resolution = self.get_resolution_value(ndvi_res, dhm[0], ndvi[0])
        # Fetch max resolution between source to ensure common_extent is bigger than max pixel size
        # Assuming square pixels
        max_res = max(dhm[0].rasterUnitsPerPixelX(), ndvi[0].rasterUnitsPerPixelX())
        # Fetch biggest tile to iterate on and fetch intersecting smaller tiles
        biggest = self.fetch_biggest_tiles(dhm, ndvi)

        # Check for cancellation
        if feedback.isCanceled():
            return {}
        
        # Define which of the DEM or DSM have the bigger tile 
        # and handle changes in workspace and formula accordingly
        rasters, tiles, formula, workspace_raster, workspace_tile = \
            self.handle_iteration(dhm, ndvi, biggest, keep_interm_files, ndvi_res)
        
        gutils.create_folders(out_dir, ["raster"])
        if compute_sieve:
            gutils.create_folders(os.path.join(out_dir, "raster"), ["tree_cover_sieve"])
        if polygonize or close:
            gutils.create_folders(out_dir, ["vector"])
            # Use 20% of progress bar for polygonization
            multiplicator = 80

        # Step for progress bar
        # Count is an estimation because there 
        # will be tiles that are not intersecting
        count = len(tiles) 
        current = 0

        feedback.pushInfo("Creating tree cover rasters.")
        for raster in rasters:
            part = 1 # To name raster parts when clip by tile
            if isinstance(raster, str):
                raster = rutils.load_raster(raster)
            # Fetch rasters intersecting biggest raster
            intersect = rutils.fetch_intersecting_tiles(raster, tiles)
            
            if feedback.isCanceled():
                return {}
            
            for tile in intersect:
                # Assign projection [is it necessary to allow this option ?]
                if assign_proj:
                    wr.assign_projection(raster)

                # Get common extent for clipping raster and align pixels
                common_ext = rutils.get_common_extent(raster, tile)
                bb = common_ext.boundingBox()
                # Ignore tiles where common_ext < max pixel resolution
                if (bb.xMaximum() - bb.xMinimum()) >= max_res \
                    and (bb.yMaximum() - bb.yMinimum()) >= max_res:

                    # Clip and align raster and tile
                    warp, clip = \
                        self.tree_cover_preprocess(raster, tile, resolution, 
                                                   common_ext, resampling, biggest, ndvi_res, 
                                                   keep_interm_files, overwrite,
                                                   workspace_raster, workspace_tile, 
                                                   part)  

                    if feedback.isCanceled():
                        return {}
                
                    tile_name = tile.name()
                    # Compute dhm and masks if wanted
                    self.generate_tree_cover(warp, clip, formula, compute_sieve, 
                                             threshold_sieve, connectedness,
                                             overwrite, out_dir, tile_name, part)
                    
                    current += 1
                    part += 1

                # Set progress
                feedback.setProgress(int((current/count) * multiplicator))
                # Check for cancellation
                if feedback.isCanceled():
                    return {}
        
        if polygonize or close:
            if close and not polygonize:
                feedback.pushWarning("Morphological closing is toggled but not the polygonize parameter.\
                                      Therefore the vector layer will be created.")
            
            feedback.pushInfo("Transforming rasters to vector.")
            self.rasters_to_vector(overwrite, close, max_res, out_dir)
            # Set progress
            feedback.setProgress(100)

    

        return {self.OUTPUT_DIR: out_dir}

    def handle_iteration(self, dhm:list[QgsRasterLayer], ndvi:list[QgsRasterLayer], 
                         biggest:bool, keep_interm_files:bool, dem_res:bool
                         )->tuple[list[QgsRasterLayer], list[QgsRasterLayer],
                                  str, str, str]:
        """
        Define which of the NDVI or DHM as the biggest tiles for future iteration.
        Define the DHM generation formula aswell depending of which one as the bigger tile.
        If intermediary files are kept, create subfolders within NDVI and DHM directory.

        Parameters
        ---
        dhm:list[QgsRasterLayer]
            List of DHM raster
        ndvi:list[QgsRasterLayer]
            List of NDVI raster
        biggest:bool
            If True a DHm tile is bigger than a NDVI tile
        keep_interm_files:bool
            If True subfolders will be created within parent directory of NDVI and DHM
        dem_res:bool
            If True the NDVI resolution will be kept, which modify treatment order.
            This change is echoed on subfolders creation

        Return
        ---
        rasters: list[QgsRasterLayer]
            List of the bigger tiles, either DHM or NDVI
        tiles: list[QgsRasterLayer]
            List of the smaller tiles, either DHM or NDVI
        formula: str
            Formula for tree cover generation. Dependant of which one as the bigger tile
        workspace_raster: str
            Directory of the bigger tiles
        workspace_tile: str
            Directory of the smaller tiles
        """
        if biggest:
            rasters = dhm
            tiles = ndvi      
            if keep_interm_files and dem_res:
                workspace_raster = self.create_interm_folders(dhm[0], ["clipped"])
                workspace_tile = self.create_interm_folders(ndvi[0], ["aligned"])
            if keep_interm_files and dem_res == False:
                workspace_raster = self.create_interm_folders(dhm[0], ["aligned"])
                workspace_tile = self.create_interm_folders(ndvi[0], ["clipped"])
        else:
            rasters = ndvi
            tiles = dhm
            if keep_interm_files and dem_res:
                workspace_tile = self.create_interm_folders(dhm[0], ["aligned"])
                workspace_raster = self.create_interm_folders(ndvi[0], ["clipped"])
            if keep_interm_files and dem_res == False:
                workspace_tile = self.create_interm_folders(dhm[0], ["clipped"])
                workspace_raster = self.create_interm_folders(ndvi[0], ["aligned"])
        
        if keep_interm_files is False:
            workspace_raster = None
            workspace_tile = None
        
        formula = "A.astype(uint8) * B.astype(uint8)"

        return rasters, tiles, formula, workspace_raster, workspace_tile
    
    def create_interm_folders(self, file:Union[QgsRasterLayer, str], folders:list[str])->str:
        """
        Create folders in a parent file directory.

        Parameters
        ---
        input_file: QgsRasterLayer | str
            Path of a file to fetch the parent directory from.
        folders: list[str]
            Name of the folders that'll be created in the parent directory
        
        Return
        ---
        workspace: str
            Return the path of the parent directory for convenience
        """
        if isinstance(file, QgsRasterLayer):
            path = file.dataProvider().dataSourceUri()
        else:
            path = file

        workspace, _ = os.path.split(path)
        gutils.create_folders(workspace, folders)

        return workspace

    def get_resolution_value(self, ndvi_res:bool, 
                             dhm:Union[QgsRasterLayer, str], 
                             ndvi:Union[QgsRasterLayer, str])->float:
        """
        From the value of a boolean parameter determine 
        if the dem resolution or the dsm resolution should be taken
        
        Parameters
        ---
        ndvi_res: boolean
            If True NDVI resolution will be used
            Else DHM resolution will be used
        dhm: QgsRasterLayer | str
        ndvi: QgsRasterLayer | str
        
        Return
        ---
        resolution:float
            Target resolution
        """
        if isinstance(dhm, str):
            dhm = rutils.load_raster(dhm)
        if isinstance(ndvi, str):
            ndvi = rutils.load_raster(ndvi)
        
        # Assuming squared pixels
        # Round because QGIS will sometimes create non null decimal 
        # at 17th or 18th place which result in missalignement
        if ndvi_res:
            resolution = round(ndvi.rasterUnitsPerPixelX(), 4)
            resolution_2 = round(dhm.rasterUnitsPerPixelX(), 4)
        else:
            resolution = round(dhm.rasterUnitsPerPixelX(), 4)
            resolution_2 = round(ndvi.rasterUnitsPerPixelX(), 4)
        
        return resolution
    
    def fetch_biggest_tiles(self, 
                            rasters_1:list[Union[QgsRasterLayer, str]], 
                            rasters_2:list[Union[QgsRasterLayer, str]])->bool:
        """
        Compare area of two raster and return a boolean.
        
        Parameters:
        ---
        rasters_1: list[QgsRasterLayer | Str]
        rasters_2: list[QgsRasterLayer | Str]
        
        Return
        ---
        biggest:bool
            True if raster_1 > raster_2 else False
        """ 
        biggest = False
        
        max_area_1 = 0
        for raster_1 in rasters_1:
            if isinstance(raster_1, str):
                raster_1 = rutils.load_raster(raster_1)
            if raster_1.extent().area() > max_area_1:
                max_area_1 = raster_1.extent().area() 
                    
        max_area_2 = 0
        for raster_2 in rasters_2:
            if isinstance(raster_2, str):
                raster_2 = rutils.load_raster(raster_2)
            if raster_2.extent().area() > max_area_2:
                max_area_2 = raster_2.extent().area() 
                    
        
        if max_area_1 > max_area_2:
            biggest = True
        
        return biggest
    
    def check_files_exists(self, files:list[str], overwrite:bool)->None:
        """
        From a list of path, check if a files already exists.
        If owerwrite is not toggle, algorithm will end there 
        and a feedback will be pushed to the user.

        Parameters
        ---
        files: list[str]
            Path of files to check
        overwrite: bool
            Overwrite parameter of the algorithm
        Return
        ---
        None
        """
        for file in files:
            exists = gutils.is_file_exists(file)
            if exists and overwrite is False:
                _, tail = os.path.split(file)
                raise QgsProcessingException(
                    f"{tail} already exists. Please delete files/folders \
                    or toggle overwrite parameter")
            if exists and overwrite and gutils.is_file_exists(f"{file}.aux.xml"):
                os.remove(f"{file}.aux.xml")
            if exists == False:
                self.just_created.append(file)
    
    def tree_cover_preprocess(self, raster:QgsRasterLayer, tile:QgsRasterLayer, 
                       resolution:float, common_ext:QgsGeometry, resampling:int,
                       biggest:bool, ndvi_res:bool, keep_interm_files:bool, 
                       overwrite:bool, workspace_raster:str, workspace_tile:str, 
                       count:int)->tuple[str, str]:
        """
        Align and clip both DHM and NDVI accordingly with tree cover wanted resolution.
        If intermediary files should be kept 
        it'll write the various outputs in DHM or NDVI directory.

        Some case inside the method seems redo

        Parameters
        ---
        raster: QgsRasterLayer
            Either DHM or NDVI tile depending of which as the bigger ones.
        tile: QgsRasterLayer 
            Either DHM or NDVI tile depending of which as the smaller ones.
        resolution: float
            Resampling resolution for DHM or NDVI raster.
        common_ext: QgsGeometry
            Common extent between the DHM or NDVI tiles.
        resampling: int
            Gdal warp algorithm to be used
        biggest: bool
            If True the biggest tile (raster) is the DHM.
        ndvi_res: bool
            If True NDVI resolution will be applied to tree cover.
            Else DHm resolution will be applied to tree cover.
        keep_interm_files: bool
            If True clipped and warped rasters will be saved in DHM or NDVI directory.
        overwrite: bool
            Will overwrite output if it exists
        workspace_raster: str
            Directory of either the DHM or NDVI.
        workspace_tile: str
            Directory of either the DHM or NDVI.
        count: int

        Return
        ---
        warp:str
            Warped DHM or NDVI path.
        clip:str
            Clipped DHM or NDVI path.
        """ 
        clip_uri = "TEMPORARY_OUTPUT"
        warp_uri = "TEMPORARY_OUTPUT"
        # Define if intermediary files should be kept and where to save them
        # Depending of resolution wanted by user
        if (biggest and ndvi_res) or (biggest == False and ndvi_res == False): 
            if keep_interm_files:
                # DHM = raster, NDVI = tile and DHM should be warp
                # DHM = tile, NDVI = raster and NDVI should be warp
                warp_uri = os.path.join(workspace_raster, "aligned", 
                                        f"{raster.name()}.tif")
                clip_uri = os.path.join(workspace_tile, "clipped", 
                                        f"{tile.name()}.tif")
                if clip_uri in self.just_created: # In case tile is between two rasters
                    clip_uri = os.path.join(workspace_tile, "clipped", 
                                            f"{tile.name()}_{count}.tif")
                if warp_uri in self.just_created:
                    warp_uri = os.path.join(workspace_raster, "aligned", 
                                            f"{raster.name()}_{count}.tif")
                self.check_files_exists([warp_uri, clip_uri], overwrite)

            # Align and clip raster between them    
            warp = wr.gdal_warp(raster, resolution, common_ext, resampling, warp_uri, overwrite)
            clip = wr.gdal_clip_by_extent(tile, common_ext, clip_uri)
        else:
            if keep_interm_files:
                # DHM = raster, NDVI = tile and NDVI should be warp
                # DHM = tile, NDVI = raster and DHM should be warp
                clip_uri = os.path.join(workspace_raster, "clipped", 
                                        f"{raster.name()}_{count}.tif")
                warp_uri = os.path.join(workspace_tile, "aligned", 
                                        f"{tile.name()}.tif")
                if clip_uri in self.just_created: # In case tile is between two rasters
                    clip_uri = os.path.join(workspace_raster, "clipped", 
                                            f"{raster.name()}_{count}.tif")
                if warp_uri in self.just_created:
                    warp_uri = os.path.join(workspace_tile, "aligned", 
                                            f"{tile.name()}_{count}.tif")
                self.check_files_exists([clip_uri, warp_uri], overwrite)

            # Align and clip raster between them
            warp = wr.gdal_warp(tile, resolution, common_ext, resampling, warp_uri, overwrite)
            clip = wr.gdal_clip_by_extent(raster, common_ext, clip_uri)
        
        return warp, clip
    
    def generate_tree_cover(self, warp:str, clip:str, formula:str, 
                     compute_sieve:bool, threshold_sieve:float, 
                     connectedness:int, overwrite:bool, out_dir:str, 
                     name:str, count:int)->None:
        """
        Generate tree cover from a DHM and a NDVI with the same extent and resolution.
        It is also possible to sieve the tree cover 
        in this case a folder will be created in the tree cover folder.

        Parameters
        ---
        warp:str
            Path of the warped raster. Either DHM or NDVI.
        clip:str
            Path of the clipped raster. Either DHM or NDVI.
        formula:str 
            Order of substraction depending of tile size between DHM or NDVI.
        compute_sieve:bool
            If True will sieve the tree cover DHM below the threshold.
        threshold_sieve:float
            Value used to threshold the DHM.
        connectedness : int
            If 0, 4D will be applied.
            If 1, 8D will be applied.
        overwrite:bool
            Will overwrite output if it exists
        out_dir: str
            Directory where DHM will be saved.
            If the masks are created they'll be saved within a subdirectory.
        name: str
            Name of the input raster which will be used to name the created DHM
            -> DHM_name
        count: int
        
        Return
        ---
        None
        """
        tree_cover_uri = os.path.join(out_dir, "raster", f"tree_cover_fr_{name}.tif")
        if tree_cover_uri in self.just_created:
            tree_cover_uri = os.path.join(out_dir, "raster", 
                                          f"tree_cover_fr_{name}_{count}.tif")
        self.check_files_exists([tree_cover_uri], overwrite)
        tree_cover = wr.gdal_calc([(clip, 1), (warp, 1)], formula, 0,
                                   output=tree_cover_uri, overwrite=overwrite)

        if compute_sieve:
            sieve_uri = os.path.join(out_dir, "raster", "tree_cover_sieve",
                                     f"SIEVE_fr_{name}.tif")
            if sieve_uri in self.just_created:
                sieve_uri = os.path.join(out_dir, "raster", "tree_cover_sieve",
                                         f"SIEVE_fr_{name}_{count}.tif")
            self.check_files_exists([sieve_uri], overwrite)

            tree_cover = wr.gdal_sieve(tree_cover, threshold_sieve, 
                              connectedness, sieve_uri)

        self.files.append(tree_cover)
    
    def rasters_to_vector(self, overwrite:bool, close:bool,
                          max_res:float, out_dir:str)->None:
        """
        Will fetch all the tif files from out_dir 
        or out_dir/tree_cover_sieve if compute_sieve is True.
        Then it'll merge all the raster before polygonizing.
        D4 will be used to avoid creating one vertex gap, 
        which make the median axis creation bug afterward.

        Parameters
        ---
        overwrite: bool
            Will overwrite output if it exists
        close: bool
            If true will perform a morphological closing of the vector layer 
            with the size of the biggest pixel between the two inputs
        max_res: float
            Value of the biggest pixel of the two inputs.
        out_dir: str
            Path of the output directory.
        """
        merge = wr.gdal_merge(self.files, rtype=0)
        vector_uri = os.path.join(out_dir, "vector", "tree_cover.gpkg")
        self.check_files_exists([vector_uri], overwrite)
        if gutils.is_file_exists(vector_uri):
            os.remove(vector_uri)
        if close:
            result = wr.gdal_polygonize(merge)
            closed = qw.closing(result, max_res, True, 1, 1)
            _ = qw.multipart_to_singleparts(closed, output=vector_uri)
        else:
            _ = wr.gdal_polygonize(merge, output=vector_uri)

    def icon(self):
        """
        Should return a QIcon which is used for your provider inside
        the Processing toolbox.
        """
        return QIcon(":/plugins/hedge_tools/images/hedge_tools.png")

    def shortHelpString(self):
        """
        Returns a localised short helper string for the algorithm. This string
        should provide a basic description about what the algorithm does and 
        the parameters and outputs associated with it.
        """
        return self.tr("Generate a tree cover from any intersecting NDVI and DHM raster. \
                        Algorithm allow to select either the NDVI or DHM resolution \
                        for the tree cover files. \n\
                        The user can also select the resampling method to use \
                        when adjusting the resolution. Some methods yield better results \
                        for downsampling while others are better for upsampling. \
                        Beware some methods (like nearest neighbor) might create space between tiles. \n\
                        It is also possible and advised to use the sieve option to \
                        remove small objects (connected pixels) in the masks \
                        for further processing within HedgeTools. \
                        The sieve value is in pixel number. \n\
                        The connectedness parameter specify if 8 neighbooring pixels \
                        (diagonals included) should be tested for connection or only 4. \n\
                        Intermediary files used for clipping and aligning rasters between them can be kept. \n\
                        For further processing within the HedgeTools toolbox the tree cover \
                        should be polygonized. The tool offers this option with the \
                        \"polygonize tree cover rasters as one vector layer \" box.")
    
    def name(self):
        """
        Returns the algorithm name, used for identifying the algorithm. This
        string should be fixed for the algorithm, and must not be localised.
        The name should be unique within each provider. Names should contain
        lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "generatetreecover"

    def displayName(self):
        """
        Returns the translated algorithm name, which should be used for any
        user-visible display of the algorithm name.
        """
        return self.tr("3 - Generate tree cover")

    def group(self):
        """
        Returns the name of the group this algorithm belongs to. This string
        should be localised.
        """
        return self.tr("0 - Extraction [optional]")

    def groupId(self):
        """
        Returns the unique ID of the group this algorithm belongs to. This
        string should be fixed for the algorithm, and must not be localised.
        The group id should be unique within each provider. Group id should
        contain lowercase alphanumeric characters only and no spaces or other
        formatting characters.
        """
        return "extraction"
    
    def tr(self, string):
        """
        Returns a translatable string with the self.tr() function.
        """
        return QCoreApplication.translate("Processing", string)
    
    def checkParameterValues(self, parameters, context):
        
        if parameters[self.OUTPUT_DIR]  == "TEMPORARY_OUTPUT":
            return (False, "It is mandatory to provide a DHM output directory.")
        
        return (True, '')

    def createInstance(self):
        return GenerateTreeCoverProcessingAlgorithm()
                    
